/******************************************************************************

                              Copyright (c) 2012
                            Lantiq Deutschland GmbH
                     Am Campeon 3; 85579 Neubiberg, Germany

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

******************************************************************************/

/**
   \file lib_tapi_lt_common_cpe.c  Linetesting library
   The functions in this file serve as an API for the linetesting functionality.
   They are compiled into the application and provide the interface to the
   driver to control linetesting and get the results. The functions mainly take
   care about scaling the driver results into floating point precision.
   The functions here can only be used where floating point is supported.

   These functions are provided for calibration of the system at open loop
   conditions:
   - IFX_TAPI_NLT_OLCalibration
   - IFX_TAPI_NLT_OLConfig_Set

   Contains implementation of following user interface functions for
   GR.909 line testing:
   - Ifxphone_LT_GR909_Config
   - Ifxphone_LT_GR909_Start
   - Ifxphone_LT_GR909_Stop
   - Ifxphone_LT_GR909_GetResults

   Addtionally these functions for measurement of the line capacitance are
   available:
   - IFX_TAPI_NLT_Capacitance_Start
   - IFX_TAPI_NLT_Capacitance_Stop
   - IFX_TAPI_NLT_Capacitance_GetResults
*/

/* ============================= */
/* Includes                      */
/* ============================= */
/* system includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>

#include "sys_debug.h"
#include "ifx_types.h"
#include "drv_tapi_io.h"
#include "lib_tapi_lt_gr909.h"

/* ============================= */
/* Defines                       */
/* ============================= */

/* ============================= */
/* Local macros & definitions    */
/* ============================= */
DECLARE_TRACE_GROUP(TAPI);

/* ============================= */
/* Local function declaration    */
/* ============================= */
#if (defined (VMMC) && defined (TAPI_GR909))
extern IFX_void_t VMMC_LT_GR909_CalcResults(IFX_LT_GR909_RESULT_t *p_res,
                                            IFX_TAPI_GR909_RESULT_t *pTapiRes);
extern IFX_LT_GR909_CFG_t vmmc_cfg;
#endif
#if (defined (VIN_2CPE) && defined (TAPI_GR909))
extern IFX_void_t VINCPE_LT_GR909_CalcResults(IFX_LT_GR909_RESULT_t *p_res,
                                            IFX_TAPI_GR909_RESULT_t *pTapiRes);
extern IFX_LT_GR909_CFG_t vincpe_cfg;
#endif
#if (defined (DXT) && defined (TAPI_GR909))
extern IFX_void_t DXT_LT_GR909_CalcResults(IFX_LT_GR909_RESULT_t *p_res,
                                            IFX_TAPI_GR909_RESULT_t *pTapiRes);
extern IFX_LT_GR909_CFG_t dxt_cfg;
#endif

/* ============================= */
/* Local function definition     */
/* ============================= */
/**
   Get pointer to the IFX_LT_GR909_CFG_t config structure.

   \param  dev          Enum of \ref IFX_TAPI_GR909_DEV_t type.
   \param  p_cfg        Pointer to IFX_LT_GR909_CFG_t structure.

   \return
   IFX_SUCCESS / IFX_ERROR
*/
IFX_return_t ifx_nlt_GR909_GetConfigPtr (IFX_TAPI_GR909_DEV_t dev,
                                         IFX_LT_GR909_CFG_t **ppCfg)
{
   IFX_return_t ret = IFX_SUCCESS;

   /* use user config */
   switch (dev)
   {
#if (defined (VMMC) && defined (TAPI_GR909))
      case IFX_TAPI_GR909_DEV_VMMC:
         *ppCfg = &vmmc_cfg;
         break;
#endif
#if (defined (VIN_2CPE) && defined (TAPI_GR909))
      case IFX_TAPI_GR909_DEV_VINCPE:
         *ppCfg = &vincpe_cfg;
         break;
#endif
#if (defined (DXT) && defined (TAPI_GR909))
      case IFX_TAPI_GR909_DEV_DXT:
         *ppCfg = &dxt_cfg;
         break;
#endif
      default:
         *ppCfg = IFX_NULL;
         ret = IFX_ERROR;
         TRACE(TAPI, DBG_LEVEL_NORMAL,
                 ("\n\rIfxphone_LT_GR909_Config: incorrect device type\n\r"));
   }

   return ret;
}

/**
   Set measurement path according to Rmes (R1) resistor.

   \param  fd           Line file descriptor.

   \return
   IFX_SUCCESS / IFX_ERROR
*/
IFX_return_t ifx_nlt_SetRmesConf (IFX_int32_t fd)
{
   IFX_return_t ret;
   IFX_TAPI_NLT_CONFIGURATION_OL_t OlConf = {0};
   IFX_TAPI_NLT_CONFIGURATION_RMES_t RmesConf = {0};
   IFX_LT_GR909_CFG_t       *pCfg = IFX_NULL;

   /* read current open loop configuration to get device type. */
   ret = ioctl(fd, IFX_TAPI_NLT_CONFIGURATION_OL_GET,
               (IFX_int32_t) &OlConf);
   if (ret != IFX_SUCCESS)
   {
      /* In case of an error check the reason. If the error occured because
         the driver does not support the IOCTL return success to maintain
         compatiblity with driver that do not have this interface and run
         with default values in this case. */
      IFX_TAPI_Error_t error;

      memset (&error, 0, sizeof (error));

      ret = ioctl (fd, IFX_TAPI_LASTERR, (IFX_uintptr_t)&error);
      if (ret == IFX_SUCCESS &&
          (((error.nCode) >> 16) & 0x0000FFFF) == 0x6007)
      {
         return IFX_SUCCESS;
      }

      return IFX_ERROR;
   }

   ret = ifx_nlt_GR909_GetConfigPtr (OlConf.dev_type, &pCfg);
   if (ret != IFX_SUCCESS)
      return IFX_ERROR;

   if (pCfg->f_R1 > 1200000)
      RmesConf.nRmeas = IFX_TAPI_NLT_RMEAS_1_5MOHM;
   else
      RmesConf.nRmeas = IFX_TAPI_NLT_RMEAS_1MOHM;

   /* Set measurement path according to Rmes resistor. */
   ret = ioctl(fd, IFX_TAPI_NLT_CONFIGURATION_RMES_SET,
               (IFX_int32_t) &RmesConf);
   return ret;
}

/* ============================= */
/* Global function definition    */
/* ============================= */

/**
   This function does an open loop calibration needed for capacitance and
   GR.909 measurements.

   This function automatically sets the calculated calibration factors into
   the driver, so user doesn't need to do it.

   Open Loop calibration should be executed only once, when
   phone is NOT connected to the FXS port. User should store calibration factors
   in non-volatile memory and execute \ref IFX_TAPI_NLT_OLConfig_Set
   during next system start. This will save time during startup and still give
   most accurate measurement results.

   \param  fd           Line file descriptor.
   \param  nNrLoops     number of loops for calibration. 1 is sufficient.
   \param  b_euLike     IFX_TRUE  : EU like powerline frequency (50 Hz)
                        IFX_FALSE : US like power line frequency (60 Hz)
   \param  pConfig      Pointer to a struct of
                        \ref IFX_TAPI_NLT_CONFIGURATION_OL_t. Calculated
                        calibration factors are stored to this structure.

   \return Returns value as follows:
   - \ref IFX_SUCCESS: if successful
   - errorcode in case of an error

   \remarks
   This function automatically sets the calculated calibration factors into
   the driver, so user doesn't need to do it.
   Open Loop calibration should be executed only once, when the phone is NOT
   connected to the FXS port. The resulting calibration factors should then
   be stored in non-volatile memory and on next systems start the restored
   using \ref IFX_TAPI_NLT_OLConfig_Set. This will save time during startup
   and still give most accurate measurement results.
*/
IFX_int32_t IFX_TAPI_NLT_OLCalibration (
                                      IFX_int32_t fd,
                                      IFX_uint32_t nNrLoops,
                                      IFX_boolean_t b_euLike,
                                      IFX_TAPI_NLT_CONFIGURATION_OL_t *pConfig)
{
   IFX_return_t ret;
   IFX_int32_t  i;
   IFX_float_t fCalCapT2R = 0.0;
   IFX_float_t fCalCapT2G = 0.0;
   IFX_float_t fCalCapR2G = 0.0;
   IFX_float_t fCalResT2R = 1000000000000.0;
   IFX_float_t fCalResT2G = 1000000000000.0;
   IFX_float_t fCalResR2G = 1000000000000.0;
   IFX_TAPI_GR909_START_t   GR909_measStart;
   IFX_TAPI_GR909_RESULT_t  GR909_measRes;
   IFX_TAPI_NLT_CAPACITANCE_START_t CapStartArg = {0};
   IFX_TAPI_NLT_CAPACITANCE_RESULT_t CapRes = {0};
   /* used to cast int16 into uint32 without implicit sign extension */
   IFX_uint32_t value;
   IFX_float_t fResTemp;

   if ((pConfig == IFX_NULL) || (nNrLoops == 0))
      return IFX_ERROR;

   ret = ioctl(fd, IFX_TAPI_LINE_FEED_SET, IFX_TAPI_LINE_FEED_DISABLED);
   if (ret != IFX_SUCCESS)
       return IFX_ERROR;

   ret = ioctl(fd, IFX_TAPI_CALIBRATION_START, 0);
   if (ret != IFX_SUCCESS)
       return IFX_ERROR;
   sleep(1);

   /* Calibration for capacitance factors. */
   for (i=0; i<nNrLoops; i++)
   {
      ret = ioctl(fd, IFX_TAPI_NLT_CAPACITANCE_START,
                  (IFX_int32_t)&CapStartArg);
      if (ret != IFX_SUCCESS)
      {
         return IFX_ERROR;
      }
      sleep(1);

      ret = ioctl(fd, IFX_TAPI_NLT_CAPACITANCE_RESULT_GET,
               (IFX_int32_t)&CapRes);
      if (ret != IFX_SUCCESS)
         return IFX_ERROR;

      /* calculate the caoacitance factors. */
      if (CapRes.bValidLine2Gnd == IFX_TRUE)
      {
         fCalCapT2G = fCalCapT2G + (IFX_float_t)CapRes.nCapTip2Gnd / nNrLoops;
         fCalCapR2G = fCalCapR2G + (IFX_float_t)CapRes.nCapRing2Gnd / nNrLoops;
      }
      if (CapRes.bValidTip2Ring == IFX_TRUE)
      {
         fCalCapT2R = fCalCapT2R + (IFX_float_t)CapRes.nCapTip2Ring / nNrLoops;
      }
   }

   ret = ifx_nlt_SetRmesConf(fd);
   if (ret != IFX_SUCCESS)
      return IFX_ERROR;

   GR909_measStart.test_mask = IFX_TAPI_GR909_RFT;
   GR909_measStart.pl_freq = ((b_euLike == IFX_TRUE) ? IFX_TAPI_GR909_EU_50HZ :
                                                       IFX_TAPI_GR909_US_60HZ);

   /* Calibration for GR909 factors. */
   for (i=0; i<nNrLoops; i++)
   {
      /* start GR909 measurment. */
      ret = ioctl(fd, IFX_TAPI_GR909_START, (IFX_int32_t) &GR909_measStart);
      if (ret != IFX_SUCCESS)
         return IFX_ERROR;
      sleep(2);

      /* read results */
      ret = ioctl(fd, IFX_TAPI_GR909_RESULT, (IFX_int32_t)&GR909_measRes);
      if (ret != IFX_SUCCESS)
         return IFX_ERROR;
      if (GR909_measRes.valid & IFX_TAPI_GR909_RFT_EXT_VALID)
      {
         /* RFT tip to ring */
         value = (IFX_uint32_t)((IFX_uint16_t)GR909_measRes.RFT_T2R);
         fResTemp = (value >> 4) << (value & 0x000F);
         fCalResT2R = 1 / (1/fCalResT2R - 1/fResTemp/nNrLoops);

         /* RFT ring to ground */
         value = (IFX_uint32_t)((IFX_uint16_t)GR909_measRes.RFT_R2G);
         fResTemp = (value >> 4) << (value & 0x000F);
         fCalResR2G = 1 / (1/fCalResR2G - 1/fResTemp/nNrLoops);

         /* RFT tip to ground */
         value = (IFX_uint32_t)((IFX_uint16_t)GR909_measRes.RFT_T2G);
         fResTemp = (value >> 4) << (value & 0x000F);
         fCalResT2G = 1 / (1/fCalResT2G - 1/fResTemp/nNrLoops);
      }
      pConfig->dev_type = GR909_measRes.dev_type;
   }

   /* return calibration factors. */
   pConfig->fOlCapTip2Ring = fCalCapT2R;
   pConfig->fOlCapTip2Gnd = fCalCapT2G;
   pConfig->fOlCapRing2Gnd = fCalCapR2G;
   pConfig->fOlResTip2Ring = fCalResT2R;
   pConfig->fOlResTip2Gnd = fCalResT2G;
   pConfig->fOlResRing2Gnd = fCalResR2G;

   /* set calibration factors in driver, so user doesn't need to do it. */
   ret = ioctl(fd, IFX_TAPI_NLT_CONFIGURATION_OL_SET, (IFX_int32_t)pConfig);
   return ret;
}

/**
   Starts measurement of capacitance values.

   \param  fd           Line file descriptor.

   \return Returns value as follows:
   - \ref IFX_SUCCESS: if successful
   - errorcode in case of an error
*/
IFX_int32_t IFX_TAPI_NLT_Capacitance_Start (IFX_int32_t fd)
{
   IFX_return_t             ret;
   IFX_TAPI_NLT_CAPACITANCE_START_t Arg = {0};

   ret = ioctl(fd, IFX_TAPI_NLT_CAPACITANCE_START, (IFX_int32_t)&Arg);

   return ret;
}

/**
   Stops measurement of capacitance values immediately.

   \param  fd           Line file descriptor.

   \return Returns value as follows:
   - \ref IFX_SUCCESS: if successful
   - errorcode in case of an error
*/
IFX_int32_t IFX_TAPI_NLT_Capacitance_Stop (IFX_int32_t fd)
{
   IFX_return_t ret;
   IFX_TAPI_NLT_CAPACITANCE_STOP_t Arg = {0};

   ret = ioctl(fd, IFX_TAPI_LINE_MEASURE_CAPACITANCE_STOP, (IFX_int32_t)&Arg);

   return ret;
}

/**
   Get measurement results of the capacitance measurement.

   \param  fd           Line file descriptor.
   \param  pResult      Pointer to a struct of \ref IFX_NLT_Capacitance_Result_t
                        to be filled with the results.

   \return Returns value as follows:
   - \ref IFX_SUCCESS: if successful
   - errorcode in case of an error
*/
IFX_int32_t IFX_TAPI_NLT_Capacitance_GetResults (IFX_int32_t fd,
                                       IFX_NLT_Capacitance_Result_t *pResult)
{
   IFX_return_t                       ret;
   IFX_TAPI_NLT_CAPACITANCE_RESULT_t  meas_res;

   /* reset structures */
   memset (&meas_res, 0, sizeof (meas_res));
   memset (pResult, 0, sizeof (*pResult));

   /* read results */
   ret = ioctl(fd, IFX_TAPI_NLT_CAPACITANCE_RESULT_GET,
               (IFX_int32_t)&meas_res);
   if (ret != IFX_SUCCESS)
      return IFX_ERROR;

   /* line 2 ground capacitance valid? */
   if (meas_res.bValidLine2Gnd == IFX_TRUE)
   {
      pResult->b_line2gnd_valid = IFX_TRUE;
      pResult->f_tip2gnd_cap = meas_res.nCapTip2Gnd - meas_res.fOlCapTip2Gnd;
      if (pResult->f_tip2gnd_cap < 0)
         pResult->f_tip2gnd_cap = 0;
      pResult->f_ring2gnd_cap = meas_res.nCapRing2Gnd -
                                meas_res.fOlCapRing2Gnd;
      if (pResult->f_ring2gnd_cap < 0)
         pResult->f_ring2gnd_cap = 0;
   }

   /* tip to ring capacitance valid? */
   if (meas_res.bValidTip2Ring == IFX_TRUE)
   {
      pResult->b_tip2ring_valid = IFX_TRUE;
      pResult->f_tip2ring_cap = meas_res.nCapTip2Ring -
                                meas_res.fOlCapTip2Ring -
                                (pResult->f_tip2gnd_cap / 4) -
                                (pResult->f_ring2gnd_cap / 4);
      if (pResult->f_tip2ring_cap < 0)
         pResult->f_tip2ring_cap = 0;
   }

   return ret;
}

/**
   Configures the open loop calibration factors of the measurement path
   for line testing.

   \param  fd           Line file descriptor.
   \param  pConfig      Pointer to a struct of
                        \ref IFX_TAPI_NLT_CONFIGURATION_OL_t

   \return Returns value as follows:
   - \ref IFX_SUCCESS: if successful
   - errorcode in case of an error
*/
IFX_int32_t IFX_TAPI_NLT_OLConfig_Set (IFX_int32_t fd,
                                       IFX_TAPI_NLT_CONFIGURATION_OL_t *pConfig)
{
   IFX_return_t ret;

   ret = ioctl(fd, IFX_TAPI_NLT_CONFIGURATION_OL_SET, (IFX_int32_t)pConfig);

   return ret;
}

/**
   Configure system parameters (SLIC) to use for values calculation,
    e.g Voltage divider resistors.

   \param  dev          Enum of \ref IFX_TAPI_GR909_DEV_t type.
   \param  p_cfg        Pointer to IFX_LT_GR909_CFG_t structure.

   \return
   IFX_SUCCESS / IFX_ERROR

   \remark
   Usage of this function is optional. Default parameters are used otherwise.
*/
IFX_return_t Ifxphone_LT_GR909_Config (IFX_TAPI_GR909_DEV_t dev,
                                       IFX_LT_GR909_CFG_t *p_cfg)
{
   IFX_return_t ret = IFX_SUCCESS;
   IFX_LT_GR909_CFG_t *pUserCfg = IFX_NULL;

   /* check config integrity to ensure software safety */
   if ((p_cfg->f_R1 == 0.0) || (p_cfg->f_R2 == 0.0) || (p_cfg->f_R3 == 0.0))
      return IFX_ERROR;

   /* use user config */
   ret = ifx_nlt_GR909_GetConfigPtr (dev, &pUserCfg);
   if (pUserCfg != IFX_NULL)
   {
      *pUserCfg = *p_cfg;
   }
   else
   {
      TRACE(TAPI, DBG_LEVEL_NORMAL,
                 ("\n\rIfxphone_LT_GR909_Config: pUserCfg is NULL!\n\r"));
      ret = IFX_ERROR;
   }

   return ret;
}

/**
   Start a GR909 test or test sequence according to measurement mask
   set with \ref IFX_LT_GR909_MASK_t.

   \param  fd_line      Line file descriptor.
   \param  b_euLike     IFX_TRUE  : EU like powerline frequency (50 Hz)
                        IFX_FALSE : US like power line frequency (60 Hz).
   \param  meas_mask    Measurement mask set with values out
                        of \ref IFX_LT_GR909_MASK_t.
   \return
   IFX_SUCCESS / IFX_ERROR
*/
IFX_return_t Ifxphone_LT_GR909_Start (IFX_int32_t   fd_line,
                                      IFX_boolean_t b_euLike,
                                      IFX_uint32_t  meas_mask)
{
   IFX_return_t             ret;
   IFX_TAPI_GR909_START_t   meas_start;
   IFX_uint32_t             *mask = &meas_start.test_mask;

   ret = ioctl(fd_line, IFX_TAPI_LINE_FEED_SET, IFX_TAPI_LINE_FEED_DISABLED);
   if (ret != IFX_SUCCESS)
       return IFX_ERROR;

   ret = ifx_nlt_SetRmesConf(fd_line);
   if (ret != IFX_SUCCESS)
      return IFX_ERROR;

   memset (&meas_start, 0, sizeof (meas_start));

   /* setup powerline frequency  */
   meas_start.pl_freq = ((b_euLike == IFX_TRUE) ? IFX_TAPI_GR909_EU_50HZ :
                                                       IFX_TAPI_GR909_US_60HZ);
   /* setup driver test mask according to user mask */
   *mask |= ((meas_mask & IFX_LT_GR909_HPT_MASK)  ? IFX_TAPI_GR909_HPT  : 0);
   *mask |= ((meas_mask & IFX_LT_GR909_FEMF_MASK) ? IFX_TAPI_GR909_FEMF : 0);
   *mask |= ((meas_mask & IFX_LT_GR909_RFT_MASK)  ? IFX_TAPI_GR909_RFT  : 0);
   *mask |= ((meas_mask & IFX_LT_GR909_ROH_MASK)  ? IFX_TAPI_GR909_ROH  : 0);
   *mask |= ((meas_mask & IFX_LT_GR909_RIT_MASK)  ? IFX_TAPI_GR909_RIT  : 0);

   /* no test ? */
   if (*mask == 0)
      return IFX_ERROR;

   ret = ioctl(fd_line, IFX_TAPI_GR909_START, (IFX_int32_t) &meas_start);

   return ret;
}

/**
   Stop a GR909 test or test sequence

   \param  fd_line      Line file descriptor.

   \return
   IFX_SUCCESS / IFX_ERROR
*/
IFX_return_t Ifxphone_LT_GR909_Stop (IFX_int32_t fd_line)
{
   IFX_return_t ret;

   ret = ioctl(fd_line, IFX_TAPI_GR909_STOP, 0);

   return ret;
}

/**
   Gets Gr909 measurement results.

   \param  fd_line      Line file descriptor.
   \param  p_res        Pointer to user result structure.

   \return
   IFX_SUCCESS / IFX_ERROR

   \remark
   Only evaluate results which are marked as valid.
*/
IFX_return_t Ifxphone_LT_GR909_GetResults (IFX_int32_t            fd_line,
                                           IFX_LT_GR909_RESULT_t *p_res)
{
   IFX_return_t             ret;
   IFX_TAPI_GR909_RESULT_t  meas_res;
   IFX_void_t (*pf_calc)(IFX_LT_GR909_RESULT_t *pRes,
                                 IFX_TAPI_GR909_RESULT_t *pTapiRes) = IFX_NULL;

   /* reset structures */
   memset (&meas_res, 0, sizeof (meas_res));
   memset (p_res, 0, sizeof (*p_res));

   /* read results */
   ret = ioctl(fd_line, IFX_TAPI_GR909_RESULT, (IFX_int32_t)&meas_res);
   if (ret != IFX_SUCCESS)
      return IFX_ERROR;

   switch (meas_res.dev_type)
   {
#if (defined (VMMC) && defined (TAPI_GR909))
      case IFX_TAPI_GR909_DEV_VMMC:
         pf_calc = VMMC_LT_GR909_CalcResults;
         break;
#endif /* VMMC */

#if (defined (VIN_2CPE) && defined (TAPI_GR909))
      case IFX_TAPI_GR909_DEV_VINCPE:
         pf_calc = VINCPE_LT_GR909_CalcResults;
         break;
#endif /* VIN_2CPE */

#if (defined (DXT) && defined (TAPI_GR909))
      case IFX_TAPI_GR909_DEV_DXT:
         pf_calc = DXT_LT_GR909_CalcResults;
         break;
#endif /* DXT */

      default:
         TRACE(TAPI, DBG_LEVEL_NORMAL,
              ("\n\rIfxphone_LT_GR909_GetResults: incorrect device type\n\r"));
   }

   if (pf_calc != IFX_NULL)
   {
      pf_calc(p_res, &meas_res);
   }
   else
   {
      TRACE(TAPI, DBG_LEVEL_NORMAL,
                 ("\n\rIfxphone_LT_GR909_GetResults: pf_calc is NULL!\n\r"));
      ret = IFX_ERROR;
   }

   return ret;
}
